import sys
import os
import io
import re
import pytest
from yadupe import core

FILEPATH_1 = 'test-data/A/2.txt'
FILEPATH_1EQ = 'test-data/A/3.txt'
FILEPATH_1NEQ = 'test-data/A/a/4.txt'

TESTDATA_PATH = 'test-data'
SOURCE_1 = 'test-data/A'
SOURCE_2 = 'test-data/B'
SOURCE_3 = 'test-data/C'
SOURCE_4 = 'test-data/D'
RESULT_DIR = 'test-data/res'

DUPLICATE_NAME_1 = 'Filename: 2.txt'
DUPLICATE_NAME_2 = 'Filename: 6.txt'
DUPLICATE_NAME_3 = 'Filename: 7.txt'

DUPLICATE_METADATA_1 = 'Size: 3432 byte, last modified: 2019-08-08T09:59:19ZGMT'
DUPLICATE_METADATA_2 = 'Size: 758 byte, last modified: 2019-08-08T10:05:43ZGMT'
DUPLICATE_METADATA_3 = 'Size: 29483 byte, last modified: 2019-08-08T10:30:11ZGMT'

DUPLICATE_MOVE_TEMPLATE_1 = r'A/3.txt\s->\s(.+)/2.txt/3.txt'
DUPLICATE_MOVE_TEMPLATE_2 = r'C/a/b/d/g/10.txt\s->\s(.+)/2.txt/a/b/d/g/10.txt'
DUPLICATE_MOVE_TEMPLATE_3 = r'C/a/b/d/e/8.txt\s->\s(.+)/2.txt/a/b/d/e/8.txt'
DUPLICATE_MOVE_TEMPLATE_4 = r'B/a/c/4.txt\s->\s(.+)/6.txt/a/c/4.txt'
DUPLICATE_MOVE_TEMPLATE_5 = r'B/a/c/e/f/8.txt\s->\s(.+)/6.txt/a/c/e/f/8.txt'
DUPLICATE_MOVE_TEMPLATE_6 = r'B/a/d/3.txt\s->\s(.+)/6.txt/a/d/3.txt'
DUPLICATE_MOVE_TEMPLATE_7 = r'C/a/b/d/f/9.txt\s->\s(.+)/6.txt/a/b/d/f/9.txt'
DUPLICATE_MOVE_TEMPLATE_8 = r'C/5.txt\s->\s(.+)/7.txt/5.txt'
DUPLICATE_MOVE_TEMPLATE_9 = r'C/3.txt\s->\s(.+)/7.txt/3.txt'
DUPLICATE_MOVE_TEMPLATE_10 = r'C/a/b/d/7.txt\s->\s(.+)/7.txt/a/b/d/7.txt'
DUPLICATE_MOVE_TEMPLATE_11 = r'D/a/b/1.txt\s->\s(.+)/7.txt/a/b/1.txt'

SHOW_FIRSTLINE = 'Duplicate list:'
SHOW_NAMELINE = 'Filename: 2.txt'
SHOW_ATTRLINE = 'Size: 3432 byte, last modified: 2019-08-08T09:59:19ZGMT'
SHOW_DUPNAME_1 = '2.txt'
SHOW_DUPNAME_2 = '3.txt'
SHOW_LASTLINE = 'End of list.'


@pytest.fixture
def namedpath_1():
    return core.NamedPath('keystring1', ['first_path', 'second_path', 'third_path'])


@pytest.fixture
def namedpath_1eq():
    return core.NamedPath('keystring1', ['first_path', 'second_path', 'third_path'])


@pytest.fixture
def namedpath_2():
    return core.NamedPath('keystring2', ['first_path', 'second_path', 'third_path'])


@pytest.fixture
def namedpath_3():
    return core.NamedPath('keystring1', ['first_path', 'second_path', ])


@pytest.fixture
def namedpath_4():
    return core.NamedPath('keystring1', ['first_path', 'second_path', 'another_third_path'])


def test_namedpath_eq(namedpath_1,
                      namedpath_1eq,
                      namedpath_2,
                      namedpath_3,
                      namedpath_4):

    assert namedpath_1 == namedpath_1eq
    assert namedpath_1 != namedpath_2
    assert namedpath_1 != namedpath_3
    assert namedpath_1 != namedpath_4


def test_similarfiles_eq():
    simfile_1 = core._SimilarFiles(FILEPATH_1)
    simfile_2 = core._SimilarFiles(FILEPATH_1)

    assert simfile_1 == simfile_2

    simfile_1.add(FILEPATH_1EQ)
    simfile_2.add(FILEPATH_1EQ)

    assert simfile_1 == simfile_2

    simfile_2.add(FILEPATH_1NEQ)
    assert simfile_1 != simfile_2


def test_filedict_eq():
    fd1 = core.FilepathDict()
    core._scan_duplicates(SOURCE_1, fd1)
    fd2 = core.FilepathDict()
    core._scan_duplicates(SOURCE_1, fd2)
    assert fd1 == fd2
    fd3 = core.FilepathDict()
    core._scan_duplicates(SOURCE_2, fd3)
    assert fd1 != fd3

    n1 = fd1.duplicateslist_count()
    n2 = fd2.duplicateslist_count()
    n3 = fd3.duplicateslist_count()

    assert n1 == n2
    assert n1 == n3


def test_print(capfd):
    fd1 = core.FilepathDict()
    core._scan_duplicates(SOURCE_1, fd1)
    fd1.print_duplicates()
    out = [line for line in capfd.readouterr()[0].split('\n') if line.strip()]

    assert out[0] == SHOW_FIRSTLINE
    assert out[1] == SHOW_NAMELINE
    assert out[2] == SHOW_ATTRLINE
    assert out[3].endswith(SHOW_DUPNAME_1)
    assert out[4].endswith(SHOW_DUPNAME_2)
    assert out[-1] == SHOW_LASTLINE


def test_duplicatemove():
    settings = core.Settings(True,
                             os.path.abspath(RESULT_DIR),
                             [os.path.abspath(path) for path in [
                                 SOURCE_1, SOURCE_2, SOURCE_3, SOURCE_4]],
                             False,
                             True)
    fd = core.FilepathDict()
    for el in settings.source:
        core._scan_duplicates(os.path.abspath(el), fd)
    fd = core._move_duplicates(fd,
                               settings.source,
                               settings.dest_path,
                               settings.op_test)
    report_buffer = io.StringIO()
    fd._save_duplicates_(target=report_buffer)
    report = report_buffer.getvalue()
    report_buffer.close()

    fulltext_templates = [DUPLICATE_NAME_1, DUPLICATE_NAME_2, DUPLICATE_NAME_3,
                          DUPLICATE_METADATA_1, DUPLICATE_METADATA_2, DUPLICATE_METADATA_3]
    for template in fulltext_templates:
        if template in report:
            continue
        else:
            assert False, f'{template} not found.'

    re_templates = [DUPLICATE_MOVE_TEMPLATE_1, DUPLICATE_MOVE_TEMPLATE_2,
                    DUPLICATE_MOVE_TEMPLATE_3, DUPLICATE_MOVE_TEMPLATE_4,
                    DUPLICATE_MOVE_TEMPLATE_5, DUPLICATE_MOVE_TEMPLATE_6,
                    DUPLICATE_MOVE_TEMPLATE_7, DUPLICATE_MOVE_TEMPLATE_8,
                    DUPLICATE_MOVE_TEMPLATE_9, DUPLICATE_MOVE_TEMPLATE_10,
                    DUPLICATE_MOVE_TEMPLATE_11]
    for template in re_templates:
        if re.search(template, report):
            continue
        else:
            assert False, f'{template} not found.'


def test_purgeempty():
    path = os.path.join(os.path.abspath(TESTDATA_PATH), 'emptytest/subdir')
    os.makedirs(path, exist_ok=True)
    assert os.access(path, os.F_OK)

    core._remove_empty_subdirs([path])
    assert not os.access(path, os.F_OK)



import sys
import os
import pytest
from yadupe import argutils


CL_CORRECT_1 = '-d -p test-data/A test-data/B test-data/C test-data/D '\
    '-r test-data/res'

CORRECT_VALUE_1 = "True, test-data/res, ['test-data/A', 'test-data/B', "\
    "'test-data/C', 'test-data/D'], True, False"

CL_CORRECT_2 = '-d test-data/A test-data/B test-data/C test-data/D '\
    '-r test-data/res'

CORRECT_VALUE_2 = "True, test-data/res, ['test-data/A', 'test-data/B', "\
    "'test-data/C', 'test-data/D'], False, False"

CL_CORRECT_3 = 'test-data/A test-data/B test-data/C test-data/D '\
    '-r test-data/res/duplicates.txt'

CORRECT_VALUE_3 = "False, test-data/res/duplicates.txt, ['test-data/A', "\
    "'test-data/B', 'test-data/C', 'test-data/D'], False, False"

CL_CORRECT_4 = 'test-data/A test-data/B test-data/C test-data/D '

CORRECT_VALUE_4 = "False, None, ['test-data/A', 'test-data/B', 'test-data/C', "\
    "'test-data/D', ''], False, False"

CL_INCORRECT_1 = '-d -p test-data/A/1.png test-data/D '\
    '-r test-data/res'

ERROR_VALUE_1 = 'test-data/A/1.png: must be valid path to directory.'

CL_INCORRECT_2 = '-d -p test-data/E -r test-data/res'

ERROR_VALUE_2 = 'test-data/E: must be valid path to directory.'

CL_INCORRECT_3 = '-d -p test-data/A test-data/B test-data/C test-data/D '\
    '-r test-data/A/2.txt'

ERROR_VALUE_3 = 'test-data/A/2.txt: must be valid path to directory.'

CL_INCORRECT_4 = '-d -p test-data/A test-data/B test-data/C test-data/D '\
    '-r test-data/ress'

ERROR_VALUE_4 = 'test-data/ress: must be valid path to directory.'


def test_empty():
    with pytest.raises(SystemExit):
        argutils.parse_arguments('')


def parse_and_validate(input: str):
    args = input.split(' ')
    settings = argutils.parse_arguments(args)
    settings = argutils.verify_settings(settings)
    return settings


def test_incorrect_args():
    for i, o in zip([CL_INCORRECT_1, CL_INCORRECT_2, CL_INCORRECT_3, CL_INCORRECT_4],
                    [ERROR_VALUE_1, ERROR_VALUE_2, ERROR_VALUE_3, ERROR_VALUE_4]):
        with pytest.raises(ValueError) as exinfo:
            parse_and_validate(i)
        assert str(exinfo.value) == o


def test_correct_args():
    for i, o in zip([CL_CORRECT_1, CL_CORRECT_2, CL_CORRECT_3, CL_CORRECT_4],
                    [CORRECT_VALUE_1, CORRECT_VALUE_2, CORRECT_VALUE_3, CORRECT_VALUE_4]):
        settings = parse_and_validate(i)
        str_settings = f'{settings.op_dedup}, {settings.dest_path}, '\
            f'{settings.source}, {settings.remove_empty}, {settings.op_test}'
        assert str_settings == o



import sys
import os
import pytest
from yadupe import hashutils

NO_FILE = 'aaaaaaaa'
SIMPLE_HASH_FILE = 'test-data/A/1.png'
SIMPLE_HASH_VALUE = "SimpleKey(size=175132, lmtime='2019-08-08T10:00:21ZGMT')"
SMALL_FILE = 'test-data/A/a/4.txt'
SMALL_FILE_SIZE = 325
SMALL_FILE_HASH = 'e6763eaf9633ef7d4e0e4f8206acabe846a71db1eaaa0a84b3b'\
                  'f591fead46b3da1f191e7cfd830eddd4e02a0e09962897cf7a5'\
                  '4db3cbd2726193500090e0b310'
EMPTY_FILE = 'test-data/A/a/4e.txt'


def test_simplekey_parameters():
    with pytest.raises(FileNotFoundError):
        hashutils.get_simple_key(os.path.abspath(NO_FILE))


def test_simplekey_value():
    sk = hashutils.get_simple_key(os.path.abspath(SIMPLE_HASH_FILE))
    assert str(sk) == SIMPLE_HASH_VALUE


def test_simplekey_creating():
    sk1 = hashutils.get_simple_key(os.path.abspath(SIMPLE_HASH_FILE))
    sk2 = hashutils.SimpleKey.create(sk1.size, sk1.lmtime)
    assert sk1 == sk2


def test_chunkreading_empty():
    total = 0
    for chunk in hashutils.file_chunks(os.path.abspath(EMPTY_FILE)):
        total += len(chunk)
    assert total == 0


def test_chunkreading():
    total = 0
    for chunk in hashutils.file_chunks(os.path.abspath(SMALL_FILE)):
        total += len(chunk)
    assert total == SMALL_FILE_SIZE


def test_hashing_invalid():
    with pytest.raises(FileNotFoundError):
        hashutils.hash_file(os.path.abspath(NO_FILE))


def test_hashing_value():
    h = hashutils.hash_file(os.path.abspath(SMALL_FILE))
    assert h == SMALL_FILE_HASH



import typing
import os
import time
import io
import hashlib


_LAST_MTIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ%Z"


class SimpleKey(typing.NamedTuple):
    """ Simple key to associate with each file. """

    size: int
    lmtime: str

    def __eq__(self, other):
        return self.size == other.size and self.lmtime == other.lmtime

    @staticmethod
    def create(size: int, strtime: str):
        return SimpleKey(size, strtime)


def get_simple_key(filepath: str) -> SimpleKey:
    """ Return simple key for file based on size and latest modification time. """

    file_stat = os.stat(filepath)
    return SimpleKey(file_stat.st_size,
                     time.strftime(_LAST_MTIME_FORMAT, time.gmtime(file_stat.st_mtime)))


def file_chunks(filepath, chunksize=io.DEFAULT_BUFFER_SIZE) -> bytes:
    """ Read file by chunk """

    with open(filepath, 'rb') as openedfile:
        while True:
            data = openedfile.read(chunksize)
            if not data:
                break
            yield data


def hash_file(filepath: str) -> str:
    hash_obj = hashlib.blake2b()
    for chunk in file_chunks(filepath):
        hash_obj.update(chunk)
    return hash_obj.hexdigest()



"""
Minimal CLI application to use yadupe as standalone software.

"""

import sys
import yadupe
from yadupe import argutils, core
import tqdm


def main():

    settings = argutils.parse_arguments()
    try:
        settings = argutils.verify_settings(settings)
    except ValueError as ex:
        print(f'{ex}')
        exit()

    progress = None
    max_progress = 0
    step = 0

    def progress_reset():
        if progress:
            progress.close()

    def progress_start(count: int, title: str):
        progress_reset()
        print(title)
        global progress
        progress = tqdm.tqdm(total=100)
        global max_progress
        max_progress = 100
        global step
        step = 100 / count

    def on_item_progress():
        global max_progress
        global step
        global progress
        if max_progress < (1.9 * step):
            step = max_progress
        progress.update(step)
        max_progress -= step

    def on_scan(count: int):
        progress_start(count, 'Search for duplicates:')

    def on_move(count: int):
        progress_start(count, 'Move duplicates:')

    def on_purge():
        progress_reset()
        print(f'Remove empty folders.')

    def on_report(count):
        progress_start(count, 'Save report:')

    hooks = core.HookWrapper()
    hooks.beforescanhook = on_scan
    hooks.pathscannedhook = on_item_progress
    hooks.beforemovehook = on_move
    hooks.groupmovedhook = on_item_progress
    hooks.beforepurgehook = on_purge
    hooks.beforereporthook = on_report
    hooks.groupreportedhook = on_item_progress

    core.deduplicate(settings, hooks)
    progress_reset()

if __name__ == "__main__":
    main()


import sys
import os
import typing
import argparse
from configparser import ConfigParser
import shlex
from .core import Settings


def parse_arguments(parameter_list: str = '') -> Settings:
    """ Parse argument list, return parameters.
    """

    arg_parser = argparse.ArgumentParser(
        prog='yadupe',
        description=sys.modules['yadupe.core'].__doc__)
    arg_parser.add_argument('source', nargs='+',
                            help='Source path to search duplicated files.',
                            metavar='PATH')
    arg_parser.add_argument('-d', '--deduplicate',
                            help='Scan and remove mode. Duplicates will be moved into given " \
                                "directory.',
                            action='store_true', dest='deduplicate')
    arg_parser.add_argument('-p', '--purge',
                            help='Remove empty subdirs after duplicates move.',
                            action='store_true', dest='rem_empty')
    arg_parser.add_argument('-r', '--result',
                            help='Path to report dir (optional for default search mode) OR " \
                                "directory to move duplicated files into.',
                            metavar='PATH')
    args = arg_parser.parse_args(
        parameter_list if len(parameter_list) else None)

    return Settings(args.deduplicate,
                    args.result,
                    args.source,
                    args.rem_empty,
                    False)


def verify_settings(arguments: Settings) -> Settings:
    """ Check parameters integrity. Load arguments from config file, if exist.
    Return valid argument set or raise exception.
    """

    if not arguments.dest_path is None:
        abspath = os.path.abspath(arguments.dest_path)
    if arguments.op_dedup:
        # Arguments.dest_path must be directory path
        if arguments.dest_path is None or not os.path.isdir(abspath):
            raise ValueError(
                f'{arguments.dest_path}: must be valid path to directory.')
    else:
        if not arguments.dest_path is None:
            # Arguments.dest_path must be valid filepath for new file.
            rootname, _ = os.path.split(abspath)
            if not os.access(rootname, os.F_OK):
                raise ValueError(
                    f'{arguments.dest_path} must be valid path to file to create.')

    for source in arguments.source:
        if not os.path.isdir(os.path.abspath(source)):
            raise ValueError(f'{source}: must be valid path to directory.')
    return arguments



"""
Recursively scan one or more given directories for duplicate files. 
Found duplicates list could be saved into report or printed out in console.

Also, duplicates could be moved into destination directory in safe way, 
preserving it relative path. In this case file name is written in the report, 
as well as new path for the file.
If empty sub-directories turn up after duplicates removal, the could be 
deleted as well.

"""

import sys
import os
import typing
from .hashutils import SimpleKey, get_simple_key, hash_file
from collections import deque
from itertools import count


__all__ = ['Settings', 'NamedPath',
           'FilepathDict', 'HookWrapper', 'deduplicate']

_STUB_KEY_STRING = 'stub-key-string'


class Settings(typing.NamedTuple):
    op_dedup: bool            # do deduplication
    dest_path: str            # path to destination dir, or report file
    source: typing.List[str]  # path to scan
    remove_empty: bool        # remove empty sub folders after deduplication
    op_test: bool             # test mode: only report, no real file moving


class HookWrapper(object):
    """ Wrapper class for some useful hook function, available for progress measuring. 

    beforescanhook    - callable that will be called once at the beginning of duplicate search
                        in given source paths. The callable will be passed one argument: total 
                        number of source paths.

    pathscannedhook   - callable that will be called each time when another source path scanned.
                        The callable will be passed no arguments.

    beforemovehook    - callable that will be called once at the beginning of duplicate 
                        file moving.
                        The callable will be passed one arguments: total number of found 
                        groups of duplicates.

    groupmovedhook    - callable that will be called each time when group of duplicates
                        will be moved into destination.
                        The callable will be passed no arguments.

    beforepurgehook   - callable that will be called once before starting purge of 
                        source paths.
                        The callable will be passed no arguments.

    afterpurgedhook   - callable that will be called once after finishing purge of 
                        source paths.
                        The callable will be passed no arguments.

    beforereporthook  - callable that will be called once at the beginning of report save.
                        The callable will be passed one arguments: total number of found 
                        groups of duplicates.

    groupreportedhook - callable that wil be called each time after report another
                        group of duplicates into file.
                        The callable will be passed no arguments.
    """

    def __init__(self):
        self.beforescanhook: typing.Callable = None
        self.pathscannedhook: typing.Callable = None
        self.beforemovehook: typing.Callable = None
        self.groupmovedhook: typing.Callable = None
        self.beforepurgehook: typing.Callable = None
        self.afterpurgedhook: typing.Callable = None
        self.beforereporthook: typing.Callable = None
        self.groupreportedhook: typing.Callable = None
        self.groups_count_cache: int = 0


class NamedPath(typing.NamedTuple):
    """ File name and it's duplicate path list. """
    name: str
    path: typing.List[str]

    def __eq__(self, other):
        if self.name != other.name:
            return False
        elif len(self.path) != len(other.path):
            return False
        elif self.path != other.path:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class _SimilarFiles(object):
    """ File path dictionary-based container, indexed by file content hash. 
    First added path stored w/o hashing.
    It will be hashed only on adding second path.
    """

    def __init__(self, first_path: str):
        self._pathes_ = {_STUB_KEY_STRING: first_path}
        self._length_ = 1

    def _add_(self, key: str, value: str):
        if key in self._pathes_.keys():
            self._pathes_[key].path.append(value)
        else:
            self._pathes_[key] = NamedPath(os.path.basename(value), [value])

    def add(self, extra_path: str) -> None:
        if self._length_ == 1:
            first_path = self._pathes_.pop(_STUB_KEY_STRING)
            new_key = hash_file(first_path)
            self._add_(new_key, first_path)
            self._length_ += 1

        new_key = hash_file(extra_path)
        self._add_(new_key, extra_path)

    def duplicates(self) -> typing.Iterator[NamedPath]:
        """ Iterator, return each list of file path with equivalent hash values."""

        for key in self._pathes_.keys():
            if key == _STUB_KEY_STRING:
                continue
            if len(self._pathes_[key]) > 1:
                yield self._pathes_[key]

    def __eq__(self, other):
        if not len(self._pathes_.keys()) == len(other._pathes_.keys()):
            return False
        return self._pathes_ == other._pathes_

    def __ne__(self, other):
        return not self.__eq__(other)


class FilepathDict(dict):
    """ Customized dictionary contains pairs {file-key : file-path-info}. 
    file-key - simple file key (size + last modification time)
    file-path-info - object with multiply appropriate file path
    """

    def __setitem__(self, key, value):
        """ The very first value with associated key must be added, 
        subsequent values will be appened to exist value. """

        if not key in self:
            super().__setitem__(key, _SimilarFiles(value))
        else:
            super().__getitem__(key).add(value)

    def _save_duplicates_(self, target, hooks=HookWrapper()):
        if hooks.beforereporthook:
            hooks.beforereporthook(hooks.groups_count_cache)
        print('Duplicate list:', file=target)
        for sk in self.keys():
            for duplicates in self[sk].duplicates():
                print(f'Filename: {duplicates.name}', file=target)
                print(f'Size: {sk.size} byte, last modified: {sk.lmtime}',
                      file=target)
                for filepath in duplicates.path:
                    print(f'{filepath}', file=target)
                print('', file=target)
                if hooks.groupreportedhook:
                    hooks.groupreportedhook()
        print('End of list.', file=target)

    def print_duplicates(self, hooks=HookWrapper()):
        self._save_duplicates_(target=sys.stdout, hooks=hooks)

    def save_duplicates(self, filepath: str, hooks=HookWrapper()):
        with open(filepath, 'wt') as fileout:
            self._save_duplicates_(target=fileout, hooks=hooks)

    @staticmethod
    def _iter_len(it):
        """ Fast method to count iterable elements. """
        cnt = count()
        deque(zip(it, cnt), 0)
        return next(cnt)

    def duplicateslist_count(self):
        """ Count number of duplicate groups in files_dict """
        total_cnt = 0
        for sk in self.keys():
            total_cnt += FilepathDict._iter_len(self[sk].duplicates())
        return total_cnt


def _scan_duplicates(rootpath: str, files_dict: dict):
    """ Scan rootpath for duplicates. """

    if not os.path.isabs(rootpath):
        rootpath = os.path.abspath(rootpath)
    for dirpath, _, filenames in os.walk(rootpath):
        for filename in filenames:
            absfilename = os.path.join(dirpath, filename)
            key = get_simple_key(absfilename)
            files_dict[key] = absfilename


def _move_duplicates(files_dict: dict,
                     sources: typing.List[str],
                     dest: str,
                     testmode=False,
                     hooks=HookWrapper()):
    """ Move duplicates into new location, log operation."""

    name_check_dict = {}

    if hooks.beforemovehook:
        hooks.beforemovehook(hooks.groups_count_cache)

    for sk in files_dict.keys():
        for duplicates in files_dict[sk].duplicates():

            if duplicates.name in name_check_dict.keys():
                idx = name_check_dict[duplicates.name]
                shortname = f'{duplicates.name}_{idx}'
                name_check_dict[duplicates.name] += 1
            else:
                name_check_dict[duplicates.name] = 1
                shortname = duplicates.name
            shortname = os.path.join(dest, shortname)
            if not testmode:
                os.mkdir(shortname)

            idx = 1
            iters = iter(duplicates.path[1:])
            for filepath in iters:
                for src in sources:
                    if src == os.path.commonpath([src, filepath]):
                        destpath = os.path.relpath(filepath, src)
                        destpath = os.path.join(shortname, destpath)
                        if not testmode:
                            os.makedirs(os.path.dirname(
                                destpath), exist_ok=True)
                        # log file move operation
                        duplicates.path[idx] = f'{duplicates.path[idx]} -> {destpath}'
                        idx += 1
                        # move file
                        if not testmode:
                            os.replace(filepath, destpath)
                        break
            if hooks.groupmovedhook:
                hooks.groupmovedhook()
    return files_dict


def _remove_empty_subdirs(root=typing.List[str]):
    """ Remove empty subdirs recursively. """

    for dir in root:
        for dirpath, dirnames, files in os.walk(dir):
            if not (files or dirnames):
                os.removedirs(dirpath)


def deduplicate(settings: Settings, hooks=HookWrapper()):
    file_data_dict = FilepathDict()

    # search for duplicates
    if hooks.beforescanhook:
        hooks.beforescanhook(len(settings.source))

    for el in settings.source:
        _scan_duplicates(os.path.abspath(el), file_data_dict)
        if hooks.pathscannedhook:
            hooks.pathscannedhook()

    if hooks.beforereporthook or hooks.beforemovehook:
        hooks.groups_count_cache = file_data_dict.duplicateslist_count()

    if not settings.op_dedup:
        # report result
        if settings.dest_path:
            file_data_dict.save_duplicates(filepath=os.path.join(settings.dest_path,
                                                                 'report.txt'),
                                           hooks=hooks)
        else:
            file_data_dict.print_duplicates(hooks=hooks)
    else:
        file_data_dict = _move_duplicates(file_data_dict,
                                          settings.source,
                                          settings.dest_path,
                                          settings.op_test,
                                          hooks=hooks)

        # clean up empty sub-dirs in source
        if settings.remove_empty and not settings.op_test:
            if hooks.beforepurgehook:
                hooks.beforepurgehook()
            _remove_empty_subdirs(settings.source)
            if hooks.afterpurgedhook:
                hooks.afterpurgedhook()

        # save report
        file_data_dict.save_duplicates(filepath=os.path.join(settings.dest_path, 'report.txt'),
                                       hooks=hooks)



"""
Recursively scan one or more given directories for duplicate files. 
Found duplicates list could be saved into report or printed out in console.

Also, duplicates could be moved into destination directory in safe way, 
preserving it relative path. In this case file name is written in the report, 
as well as new path for the file.
If empty sub-directories turn up after duplicates removal, the could be 
deleted as well.

Modules:

    core.py     - core functions: search for duplicates, move files,
                    log operations.
    argutils.py - command line argument parser, settings validation.

To use package without CLI, use:
from yadupe import core

To use package with CLI or settings validation, use:
from yadupe import argutils, core

"""

__version__ = "1.0.0"

__all__ = ['core', 'argutils']



import os.path
from setuptools import setup

HERE = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(HERE, "README.md")) as fid:
    README = fid.read()

setup(
    name="yadupe",
    version="1.0.0",
    description="Recursively scan one or more given directories for duplicate files.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/yugn/yadupe",
    author="yugn",
    author_email="cbznysrnpu@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Environment :: Console",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows"
    ],
    packages=['yadupe'],
    install_requires=["tqdm"],
    entry_points={
        "console_scripts": [
            "yadupe=yadupe.__main__:main",
        ]
    },
)


import sys
import os
import re
import pytest
from yadupe import core

SOURCE_1 = 'test-data/A'
SOURCE_2 = 'test-data/B'
SOURCE_3 = 'test-data/C'
SOURCE_4 = 'test-data/D'
RESULT_DIR = 'test-data/res'

DUPLICATE_NAME_1 = 'Filename: 2.txt'
DUPLICATE_NAME_2 = 'Filename: 6.txt'
DUPLICATE_NAME_3 = 'Filename: 7.txt'

DUPLICATE_METADATA_1 = 'Size: 3432 byte, last modified: 2019-08-08T09:59:19ZGMT'
DUPLICATE_METADATA_2 = 'Size: 758 byte, last modified: 2019-08-08T10:05:43ZGMT'
DUPLICATE_METADATA_3 = 'Size: 29483 byte, last modified: 2019-08-08T10:30:11ZGMT'

DUPLICATE_MOVE_TEMPLATE_1 = r'A/3.txt\s->\s(.+)/2.txt/3.txt'
DUPLICATE_MOVE_TEMPLATE_2 = r'C/a/b/d/g/10.txt\s->\s(.+)/2.txt/a/b/d/g/10.txt'
DUPLICATE_MOVE_TEMPLATE_3 = r'C/a/b/d/e/8.txt\s->\s(.+)/2.txt/a/b/d/e/8.txt'
DUPLICATE_MOVE_TEMPLATE_4 = r'B/a/c/4.txt\s->\s(.+)/6.txt/a/c/4.txt'
DUPLICATE_MOVE_TEMPLATE_5 = r'B/a/c/e/f/8.txt\s->\s(.+)/6.txt/a/c/e/f/8.txt'
DUPLICATE_MOVE_TEMPLATE_6 = r'B/a/d/3.txt\s->\s(.+)/6.txt/a/d/3.txt'
DUPLICATE_MOVE_TEMPLATE_7 = r'C/a/b/d/f/9.txt\s->\s(.+)/6.txt/a/b/d/f/9.txt'
DUPLICATE_MOVE_TEMPLATE_8 = r'C/5.txt\s->\s(.+)/7.txt/5.txt'
DUPLICATE_MOVE_TEMPLATE_9 = r'C/3.txt\s->\s(.+)/7.txt/3.txt'
DUPLICATE_MOVE_TEMPLATE_10 = r'C/a/b/d/7.txt\s->\s(.+)/7.txt/a/b/d/7.txt'
DUPLICATE_MOVE_TEMPLATE_11 = r'D/a/b/1.txt\s->\s(.+)/7.txt/a/b/1.txt'


def test_duplicatemove():
    settings = core.Settings(True,
                             os.path.abspath(RESULT_DIR),
                             [os.path.abspath(path) for path in [
                                 SOURCE_1, SOURCE_2, SOURCE_3, SOURCE_4]],
                             False,
                             True)

    core.deduplicate(settings)

    report = ''
    with open(os.path.abspath(os.path.join(RESULT_DIR, 'report.txt')), 'r') as file:
        report = file.read()

    fulltext_templates = [DUPLICATE_NAME_1, DUPLICATE_NAME_2, DUPLICATE_NAME_3,
                          DUPLICATE_METADATA_1, DUPLICATE_METADATA_2, DUPLICATE_METADATA_3]
    for template in fulltext_templates:
        if template in report:
            continue
        else:
            assert False, f'{template} not found.'

    re_templates = [DUPLICATE_MOVE_TEMPLATE_1, DUPLICATE_MOVE_TEMPLATE_2,
                    DUPLICATE_MOVE_TEMPLATE_3, DUPLICATE_MOVE_TEMPLATE_4,
                    DUPLICATE_MOVE_TEMPLATE_5, DUPLICATE_MOVE_TEMPLATE_6,
                    DUPLICATE_MOVE_TEMPLATE_7, DUPLICATE_MOVE_TEMPLATE_8,
                    DUPLICATE_MOVE_TEMPLATE_9, DUPLICATE_MOVE_TEMPLATE_10,
                    DUPLICATE_MOVE_TEMPLATE_11]
    for template in re_templates:
        if re.search(template, report):
            continue
        else:
            assert False, f'{template} not found.'
